import logging
import tempfile
from functools import cached_property
from http import HTTPStatus
from multiprocessing.pool import ThreadPool
from pathlib import Path

import flask
import werkzeug

from overhave.db import TestRunStatus
from overhave.entities.converters import ProcessingContext, get_context_by_test_run_id
from overhave.entities.report_manager import ReportManager
from overhave.entities.settings import OverhaveFileSettings, ProcessorSettings
from overhave.entities.stash import IStashProjectManager
from overhave.processing.abstract import IProcessor
from overhave.scenario import FileManager
from overhave.storage import ITestRunStorage, UniqueDraftCreationError, add_pr_url, save_draft
from overhave.testing import ConfigInjector, PytestRunner

logger = logging.getLogger(__name__)


class Processor(IProcessor):
    """ Class for application entity tasks processing in instance of :class:`ThreadPool`. """

    def __init__(
        self,
        settings: ProcessorSettings,
        file_settings: OverhaveFileSettings,
        test_run_storage: ITestRunStorage,
        injector: ConfigInjector,
        file_manager: FileManager,
        test_runner: PytestRunner,
        stash_manager: IStashProjectManager,
        report_manager: ReportManager,
    ):
        self._settings = settings
        self._file_settings = file_settings
        self._test_run_storage = test_run_storage
        self._injector = injector
        self._file_manager = file_manager
        self._test_runner = test_runner
        self._stash_manager = stash_manager
        self._report_manager = report_manager

    @cached_property
    def _thread_pool(self) -> ThreadPool:
        return ThreadPool(processes=self._settings.processes_num)

    def _run_test(self, context: ProcessingContext, alluredir: Path) -> int:
        with self._file_manager.tmp_feature_file(context=context) as feature_file:
            with self._file_manager.tmp_fixture_file(context=context, feature_file=feature_file) as fixture_file:
                return self._test_runner.run(fixture_file=fixture_file.name, alluredir=alluredir.as_posix())

    def _process_run(self, run_id: int) -> None:
        self._test_run_storage.set_run_status(run_id=run_id, status=TestRunStatus.RUNNING)
        ctx = get_context_by_test_run_id(run_id)
        results_dir = Path(tempfile.mkdtemp())
        logger.debug("Allure results directory path: %s", results_dir.as_posix())
        try:
            test_return_code = self._run_test(context=ctx, alluredir=results_dir)
        except Exception as e:
            logger.exception("Error!")
            self._test_run_storage.set_run_status(run_id=run_id, status=TestRunStatus.INTERNAL_ERROR, traceback=str(e))
            return

        logger.debug("Test returncode: %s", test_return_code)
        if test_return_code == 0:
            self._test_run_storage.set_run_status(run_id=run_id, status=TestRunStatus.SUCCESS)
        else:
            self._test_run_storage.set_run_status(
                run_id=run_id, status=TestRunStatus.FAILED, traceback="Test run failed!"
            )
        self._report_manager.create_allure_report(test_run_id=run_id, results_dir=results_dir)

    def execute_test(self, scenario_id: int, executed_by: str) -> werkzeug.Response:
        test_run_id = self._test_run_storage.create_test_run(scenario_id=scenario_id, executed_by=executed_by)
        self._thread_pool.apply_async(self._process_run, args=(test_run_id,))
        logger.debug("Redirect to TestRun details view with test_run_id %s", test_run_id)
        return flask.redirect(flask.url_for("testrun.details_view", id=test_run_id))

    def _create_version(self, test_run_id: int, draft_id: int) -> None:
        try:
            response = self._stash_manager.create_pull_request(test_run_id)
            logger.info("Stash PR response has been gotten")
            add_pr_url(draft_id=draft_id, response=response)
            logger.info("Stash PR result has been written")
        except Exception:
            logger.exception("Error while trying to create Stash PR!")

    def create_version(self, test_run_id: int, published_by: str) -> werkzeug.Response:
        try:
            draft_id = save_draft(test_run_id=test_run_id, published_by=published_by)
        except UniqueDraftCreationError:
            error_msg = "Error while creation draft!"
            logger.exception(error_msg)
            flask.flash(error_msg, category="error")
            return flask.redirect(
                flask.url_for("testrun.details_view", id=test_run_id), code=HTTPStatus.UNPROCESSABLE_ENTITY
            )
        self._thread_pool.apply_async(self._create_version, args=(test_run_id, draft_id))
        return flask.redirect(flask.url_for("draft.details_view", id=draft_id))



class BaseStashManagerException(Exception):
    """ Base exception for :class:`StashManager`. """


class StashPrCreationError(BaseStashManagerException):
    """ Exception for pull-request creation error. """


class NotSpecifiedFeatureTypeError(BaseStashManagerException):
    """ Exception for incorrect feature type error. """



import abc

from overhave.entities import Emulator, IAdminAuthorizationManager, IFeatureExtractor, ReportManager
from overhave.factory.context import OverhaveContext
from overhave.processing import IProcessor
from overhave.scenario import FileManager
from overhave.storage import IEmulationStorage, IFeatureTypeStorage
from overhave.testing import ConfigInjector, PluginResolver, PytestRunner
from overhave.transport import RedisProducer, S3Manager


class IOverhaveFactory(abc.ABC):
    """ Factory interface for application entities resolution and usage. """

    @abc.abstractmethod
    def set_context(self, context: OverhaveContext) -> None:
        pass

    @property
    @abc.abstractmethod
    def context(self) -> OverhaveContext:
        pass

    @property
    @abc.abstractmethod
    def test_runner(self) -> PytestRunner:
        pass

    @property
    @abc.abstractmethod
    def feature_extractor(self) -> IFeatureExtractor:
        pass

    @property
    @abc.abstractmethod
    def file_manager(self) -> FileManager:
        pass

    @property
    @abc.abstractmethod
    def injector(self) -> ConfigInjector:
        pass

    @property
    @abc.abstractmethod
    def processor(self) -> IProcessor:
        pass

    @property
    @abc.abstractmethod
    def auth_manager(self) -> IAdminAuthorizationManager:
        pass

    @property
    @abc.abstractmethod
    def feature_type_storage(self) -> IFeatureTypeStorage:
        pass

    @property
    @abc.abstractmethod
    def redis_producer(self) -> RedisProducer:
        pass

    @property
    @abc.abstractmethod
    def emulation_storage(self) -> IEmulationStorage:
        pass

    @property
    @abc.abstractmethod
    def emulator(self) -> Emulator:
        pass

    @property
    @abc.abstractmethod
    def s3_manager(self) -> S3Manager:
        pass

    @property
    @abc.abstractmethod
    def report_manager(self) -> ReportManager:
        pass

    @property
    @abc.abstractmethod
    def plugin_resolver(self) -> PluginResolver:
        pass



import logging

from flask_login import current_user
from wtforms import Form, ValidationError

from overhave import db
from overhave.admin.views.base import ModelViewConfigured

logger = logging.getLogger(__name__)


class TagsView(ModelViewConfigured):
    """ View for :class:`Feature` table. """

    can_view_details = False

    column_list = (
        "id",
        "value",
        "created_at",
        "created_by",
    )

    form_excluded_columns = ("created_at",)

    def on_model_change(self, form: Form, model: db.Tags, is_created: bool) -> None:
        if not is_created:
            if current_user.login == model.created_by or current_user.role == db.Role.admin:
                return
            raise ValidationError("Only tag creator or administrator could edit it!")
        model.created_by = current_user.login

    def on_model_delete(self, model: db.Tags) -> None:
        if not (current_user.login == model.created_by or current_user.role == db.Role.admin):
            raise ValidationError("Only author or administrator could delete tags!")



import logging
from pathlib import Path
from typing import Dict, List, Optional, Set

from overhave.entities import OverhaveFileSettings

logger = logging.getLogger(__name__)


class IncorrectWorkDirError(Exception):
    """ Exception for situation with ValueError while resolving plugin relative path to workdir. """


class PluginResolver:
    """ Class for custom pytest-bdd steps modules resolution. """

    def __init__(self, file_settings: OverhaveFileSettings):
        self._file_settings = file_settings
        self._plugins = self._resolve_plugins(directory=self._file_settings.steps_dir)
        logger.debug("Available Overhave pytest plugin modules: %s", self._plugins)

    @staticmethod
    def _check_dir_compliance(path: Path) -> bool:
        return all((path.is_dir(), not path.name.startswith("."), not path.name.startswith("_")))

    @staticmethod
    def _check_module_compliance(path: Path) -> bool:
        return all(
            (not path.is_dir(), path.name.endswith(".py"), not path.name.startswith("."), not path.name.startswith("_"))
        )

    @classmethod
    def _extract_recursively(cls, folder: Path) -> List[Path]:
        modules = []
        for path in folder.iterdir():
            if cls._check_dir_compliance(path):
                submodules = cls._extract_recursively(path)
                modules.extend(submodules)
                continue
            if not cls._check_module_compliance(path):
                continue
            modules.append(path)
        return modules

    @classmethod
    def _resolve_plugins(cls, directory: Path) -> Dict[str, List[Path]]:
        plugin_folders = [d for d in directory.iterdir() if cls._check_dir_compliance(d)]
        plugin_dict = {folder.name: cls._extract_recursively(folder) for folder in plugin_folders}

        plugin_modules = [m for m in directory.iterdir() if cls._check_module_compliance(m)]
        for folder in plugin_dict:
            plugin_dict[folder].extend(plugin_modules)
        return plugin_dict

    def _format_path(self, path: Path) -> str:
        try:
            relative_path = path.relative_to(self._file_settings.workdir).as_posix()
            return relative_path.replace("/", ".").rstrip(".py")
        except ValueError as e:
            raise IncorrectWorkDirError("Could not resolve plugin relative path!") from e

    def get_plugins(self, plugin_type: Optional[str] = None) -> List[str]:
        if isinstance(plugin_type, str):
            if self._plugins.get(plugin_type) is None:
                raise KeyError(f"Specified plugin type '{plugin_type}' does not exist!")
            plugins = [self._format_path(x) for x in self._plugins[plugin_type]]
            logger.debug("Return plugins by plugin_type='%s': %s", plugin_type, plugins)
            return plugins

        joined_modules: Set[str] = set()
        for key in self._plugins:
            joined_modules.update(set(self.get_plugins(key)))
        modules_list = list(joined_modules)
        logger.debug("Return all existing plugins: %s", modules_list)
        return modules_list



# flake8: noqa
from .emulation import EmulationStorage, EmulationStorageError, IEmulationStorage
from .feature_type import FeatureTypeStorage, IFeatureTypeStorage
from .test_run import ITestRunStorage, TestRunStorage
from .version import UniqueDraftCreationError, add_pr_url, get_last_draft, get_previous_draft, save_draft
